import time
import logging
from typing import Tuple, List

import pandas as pd
import numerapi
import tqdm

from numerai_reports import utils


napi = numerapi.NumerAPI(verbosity='warn')
logger = logging.getLogger(__name__)


def fetch_models(models: List[str],
                 batch_size: int = 100) -> Tuple[pd.DataFrame, pd.DataFrame]:
    dfs = []
    medals = []
    for chunk in tqdm.tqdm(utils.chunks(models, batch_size)):
        model_queries = [f'_{i} : v3UserProfile(modelName : "{m}"){{...f}}'
                         for i, m in enumerate(chunk)]
        query = '{' + '\n'.join(model_queries) + """}
            fragment f on V3UserProfile {
                username
                id
                medals {
                  gold
                  silver
                  bronze
                }
                roundModelPerformances {
                  corr
                  roundResolveTime
                  fnc
                  mmc
                  corrWMetamodel
                  payout
                  roundNumber
                  selectedStakeValue
                }
            }"""
        try:
            raw = napi.raw_query(query)['data']
        except KeyError:
            logger.error("failed fetching chunk, retrying..")
            time.sleep(60)
            raw = napi.raw_query(query)['data']
        for _, vals in raw.items():
            if vals is None:
                continue
            df = pd.DataFrame(vals['roundModelPerformances'])
            df['model'] = vals['username']
            df['id'] = vals['id']
            vals['medals']['model'] = vals['username']
            dfs.append(df)
            medals.append(vals['medals'])

    # put everything together
    df = pd.concat(dfs)
    medals = pd.DataFrame.from_records(medals)
    return df, medals


def fetch_leaderboard(limit: int = 10000) -> pd.DataFrame:
    query = '''
        query($limit: Int
              $offset: Int) {
          v2Leaderboard(limit: $limit
                        offset: $offset) {
            username
            mmcRep
            corrRep
            fncRep
            rank
            fncRank
            mmcRank
          }
        }
    '''
    arguments = {'limit': limit, 'offset': 0}
    raw = napi.raw_query(query, arguments)
    df = pd.DataFrame(raw['data']['v2Leaderboard'])
    df.rename(columns={'username': 'model',
                       'rank': 'corr_rank'}, inplace=True)
    df.columns = [utils.to_snake_case(col) for col in df.columns]
    return df


def fetch_from_api(limit: int = 10000) -> Tuple[pd.DataFrame, pd.DataFrame]:
    leaderboard = fetch_leaderboard(limit=limit)
    df, medals = fetch_models(leaderboard['model'].to_list())
    df.rename(columns={'corrWMetamodel': "corr_with_mm",
                       'roundNumber': 'round',
                       'selectedStakeValue': 'stake',
                       "roundResolveTime": "date"}, inplace=True)
    df.columns = [utils.to_snake_case(col) for col in df.columns]
    df.dropna(subset=["corr"], inplace=True)
    df['stake'] = df['stake'].astype("float")
    df['date'] = pd.to_datetime(df['date']).dt.date

    df.drop(columns=['id'], inplace=True)

    leaderboard = leaderboard.merge(medals, on="model", how="left")

    active_rounds = df.groupby("model", as_index=False)["round"].count()
    active_rounds.rename(columns={"round": "active_rounds"}, inplace=True)
    leaderboard = leaderboard.merge(active_rounds, on="model", how="left")
    df.dropna(subset=['corr'], inplace=True)

    return df, leaderboard


class Data(metaclass=utils.Singleton):

    def __init__(self):
        self._details = None
        self._leaderboard = None

    @property
    def leaderboard(self):
        if self._leaderboard is None:
            self.load()
        return self._leaderboard

    @property
    def details(self):
        if self._details is None:
            self.load()
        return self._details

    def load(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        details, leaderboard = fetch_from_api()
        self._details = details
        self._leaderboard = leaderboard
        return details, leaderboard


if __name__ == "__main__":
    df_details, df_leaderboard = fetch_from_api(limit=100)
    print(df_details)
    print(df_leaderboard)



import re


def to_snake_case(name: str) -> str:
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(
                Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



import logging

logging.basicConfig()


changes = {
    # FIXME round number wrong
    #
    "staking_only": 100,
    # FIXME
    "reputation_bonus_0.1": 101,
    "staking_bonus_start": 154,
    "reputation_bonus_start": 158,
    "reputation_bonus_averaging": 164,
    "correlation_score_start": 168,
    }



import tabulate


import data
import reports
import leaderboard


mails = {
         151: {'nmr_total': 0.12, 'usd_total': 2.22},
         152: {'nmr_total': 1.97, 'usd_total': 5.87},
         153: {'nmr_total': -0.05, 'usd_total': 0.0,
               'nmr_staked': 1.3, 'nmr_rep_bonus': 0.2,
               'nmr_burned': 0.25},
         154: {'nmr_total': -0.42, 'usd_total': 0.28,
               'nmr_rep_bonus': 0.1, 'nmr_staked': 2.6,
               'nmr_burned': 0.59},
         155: {'nmr_total': -0.21, 'usd_total': 1.1},
         156: {'nmr_total': -0.62, 'usd_total': 0.89},
         157: {'nmr_total': -1.17, 'usd_total': 0.37,
               'nmr_burned': 1.36, 'nmr_rep_bonus': 0.1},
         158: {'nmr_total': -0.68, 'usd_total': 0},
         159: {'nmr_total': -1.16, 'usd_total': 0},
         160: {'nmr_total': -0.23, 'usd_total': 0.28},
         161: {'nmr_total': 0.66, 'usd_total': 2.32},
         162: {'nmr_total': 0.93, 'usd_total': 3.95},
         163: {'nmr_total': 0.8, 'usd_total': 3.38},
         164: {'nmr_total': 0.48, 'usd_total': 2.15},
         165: {'nmr_total': 0.39, 'usd_total': 1.66},
         166: {'nmr_total': 0.48, 'usd_total': 2.14},

        }

for round_num, values in mails.items():
    df = reports.payments('uuazed3', round_num)
    results = []
    count = 0

    for metric, value in values.items():
        report_val = df.loc[round_num][metric]
        count += report_val != value
        results.append([metric, value, report_val, report_val == value])

    if count > 0:
        print(f"======== {round_num} =======")
        print(df)
        print(tabulate.tabulate(results, headers=['metric', 'target', 'numerai_reports', 'same']))


df = leaderboard.Leaderboard()[157]
print(df[df.username=="uuazed3"])



from typing import List
import logging

import pandas as pd

from numerai_reports.data import Data


logger = logging.getLogger(__name__)


def all_positive(rounds: int = 20, metric: str = "corr") -> pd.DataFrame:
    "models having positive 'metric' in the each of the last 'rounds' rounds"
    if metric not in {'corr', 'fnc', 'mmc'}:
        raise ValueError(f"invalid metric {metric}")
    df = Data().details
    subset = df[df['round'] > df['round'].max() - rounds].copy()
    subset['success'] = subset[metric] >= 0

    grouped = subset.groupby('model').agg({"success": "sum", metric: "mean"})
    grouped = grouped[grouped['success'] == rounds]
    grouped.sort_values(metric, ascending=False, inplace=True)
    grouped.rename(columns={metric: f'mean_{metric}'}, inplace=True)
    return grouped[f'mean_{metric}']


def all_star_club(top_n: int = 100) -> pd.DataFrame:
    cols = ['model', 'account', "corr_rank", "corr_rep",
            "mmc_rank", "mmc_rep", "fnc_rank", "fnc_rep"]
    lb = Data().leaderboard
    res = lb.loc[
                (lb['corr_rank'] <= top_n) &
                (lb['mmc_rank'] <= top_n) &
                (lb['fnc_rank'] <= top_n)][cols].copy()
    return res


def payouts(models: List[str], groupby: str = "round") -> pd.DataFrame:
    """aggregated payouts for a list of models"""
    df = Data().details
    subset = df[df['model'].isin(models)].copy()
    result = subset.groupby(groupby)['payout'].sum()
    return result


def medals_leaderboard(limit: int = 10,
                       orderby: str = "total") -> pd.DataFrame:
    cols = ['gold', 'silver', 'bronze', 'active_rounds']
    lb = Data().leaderboard[cols]
    lb.insert(3, "total", lb[['gold', 'silver', 'bronze']].sum(axis=1))
    lb.sort_values(orderby, ascending=False, inplace=True)
    return lb.astype(int).fillna(0).head(limit)


def models_of_account(model: str) -> List[str]:
    """return all models that belong the same account given a model name"""
    lb = Data().leaderboard
    try:
        account = lb[lb['model'] == model]['account'].iloc[0]
    except IndexError:
        logger.warning(f"Unknown model {model}. Check spelling.")
        return []
    res = lb[lb['account'] == account]['model'].unique()
    return list(res)




import logging
logging.basicConfig()
logging.getLogger().setLevel(logging.DEBUG)


import data
import reports



df = reports.retroactive_stake_bonus()
print(df)
exit


reports.all_star_club(162)
reports.out_of_n(160, 162)
reports.summary(162, 163)
reports.dominance("uuazed", 162)
reports.friends("uuazed", 162)
reports.reputation_bonus(162)
reports.reputation("uuazed", 162)
reports.pass_rate(162)


#print(reports.all_star_club(160))

#exit()

#lb = data.fetch_leaderboard(142, 161)
#print(reports.friends(lb, "uuazed"))

#exit(

start = 150
end = 163


print("uuazed")
print(reports.payments(['uuazed'], start, end))
print("uuazed2")
print(reports.payments(['uuazed2'], start, end))
print("uuazed3")
print(reports.payments(['uuazed3'], start, end))
print("uuazed account")
print(reports.payments(['uuazed', 'uuazed2', 'uuazed3'], start, end))
print("anna accounts")
print(reports.payments(['anna1', 'anna2', 'anna3'], start, end))
print("all")
print(reports.payments(['uuazed', 'uuazed2', 'uuazed3', 'anna1', 'anna2', 'anna3'], start, end))


reports.reputation_bonus(end).to_csv("rep.csv")



from setuptools import setup
from setuptools import find_packages


def load(path):
    return open(path, 'r').read()


numerai_reports_version = '0.5.5'


classifiers = [
    "Development Status :: 3 - Alpha",
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Topic :: Scientific/Engineering"]


if __name__ == "__main__":
    setup(
        name="numerai_reports",
        version=numerai_reports_version,
        maintainer="uuazed",
        maintainer_email="uuazed@gmail.com",
        description="Reports for the Numerai machine learning competition",
        long_description=load('README.md'),
        long_description_content_type='text/markdown',
        url='https://github.com/uuazed/numerai_reports',
        platforms="OS Independent",
        classifiers=classifiers,
        license='MIT License',
        package_data={'numerai_reports': ['LICENSE', 'README.md']},
        packages=find_packages(exclude=['tests']),
        install_requires=["pandas", "numerapi", "pyarrow", "fsspec", "gcsfs"],
        )
